/*
 * ALSA Patch Bay
 *
 * Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *
 * You have permission to use this file under the GNU General
 * Public License, version 2 or later.  See the file COPYING
 * for the full text.
 *
 */


#include "apb-gtkmm.h"

namespace APB {
namespace GTKmm {


////////////////////////////////
//////// ChoiceWidget //////////
////////////////////////////////

Gtk::RadioButton::Group ChoiceWidget::_group;

void ChoiceWidget::setInput () {
  if (_spinButton) _spinButton->set_sensitive (_toggleButton->get_active());
}

ChoiceWidget::ChoiceWidget (ChoiceType type,
                            const Choice& choice)
                                // FIXME: these shouldn't be restricted
  : _spinButtonAdj (choice._value, -10000.0, 10000.0) {
  if (type == ANY) {
    _toggleButton = manage (new Gtk::CheckButton (choice._choice));
  } else {
    _toggleButton = manage (new Gtk::RadioButton (_group, choice._choice));
  }
  pack_start (*_toggleButton);
  
  if (choice._takesInt) {
    _spinButton = manage (new Gtk::SpinButton (_spinButtonAdj));
    _spinButton->set_numeric ();
    this->setInput ();
    _toggleButton->signal_toggled().connect (SigC::slot (*this, &ChoiceWidget::setInput));
    pack_end (*_spinButton);
  } else {
    _spinButton = 0;
  }

  show_all ();
}


const ChoiceResult& ChoiceWidget::getResult () const {
  static ChoiceResult result;
  result.first = _toggleButton->get_active ();
  if (_spinButton) {
    result.second = _spinButton->get_value_as_int ();
  } else {
    result.second = 0;
  }
  return result;
}

ChoiceWindow::ChoiceWindow (const string& title, const ChoiceSpec& choiceSpec)
  : Gtk::Dialog (title, true) {
  Gtk::VBox * vbox = get_vbox ();
  Gtk::Label * message = manage (new Gtk::Label (choiceSpec._message));
  vbox->pack_start (*message);
  for (list<Choice>::const_iterator it = choiceSpec._choices.begin();
       it != choiceSpec._choices.end ();
       ++it) {
    ChoiceWidget * choiceWidget = manage (new ChoiceWidget (choiceSpec._type, *it));
    _choiceWidgets.push_back (choiceWidget);
    vbox->pack_start (*choiceWidget);
  }
  Gtk::Button * okButton = manage (new Gtk::Button(Gtk::StockID(GTK_STOCK_OK)));
  add_action_widget (*okButton, 0);
  show_all ();
}

vector<ChoiceResult> ChoiceWindow::getChoice () {
  run ();

  vector<ChoiceResult> results;
  for (list<ChoiceWidget *>::iterator it = _choiceWidgets.begin ();
       it != _choiceWidgets.end ();
       ++it) {
    results.push_back ((*it)->getResult ());
  }

  return results;
}

//////////////////////////
//////// LogBox //////////
//////////////////////////

void LogBox::log (DebugLevel level, const string& message) {
  if (level > debugLevel) return;
  cout << message << endl;
  if (!_textBuffer) return;
  ustring l (message + "\n" + _textBuffer->get_text(_textBuffer->begin(),
                                  _textBuffer->end(), true));
  _textBuffer->set_text (l);
}


/////////////////////////////////
/////////// Colour //////////////
/////////////////////////////////

Colour Colour::_currentColour (ustring("white"));

const Colour& Colour::getRandomColour () {
  static Colour colour;
  colour.set_red (Misc::randomNumber(0, 65535));
  colour.set_green (Misc::randomNumber(0, 65535));
  colour.set_blue (Misc::randomNumber(0, 65535));
  return colour;
}


const Colour& Colour::getNextColour () {
  int i = _currentColour.get_red ();
  if (i == 0) {
    i = _currentColour.get_green ();
    if (i == 0) {
      i = _currentColour.get_blue ();
      if (i == 0) {
        _currentColour.set ("white");
      } else {
        i -= _incr;
        if (i < 0) i = 0;
        _currentColour.set_blue (i);
      }
    } else {
      i -= _incr;
      if (i < 0) i = 0;
      _currentColour.set_green (i);
    }
  } else {
    i -= _incr;
    if (i < 0) i = 0;
    _currentColour.set_red (i);
  }
  return _currentColour;
}


///////////////////////////////////
//////// ButtonArray //////////////
///////////////////////////////////

Colour& ButtonArray::getColour (int index) {
  int lsize = _lineColours.size();
  if (index < lsize) {
    return _lineColours[index];
  }
  for (int i = lsize-1; i <= index; ++i) {
    _lineColours.push_back (Colour::getRandomColour());
  }
  return _lineColours[index];
}

ButtonArray::ButtonArray (APB::Driver * driver, APB::UI * ui)
  : _driver (driver),
    _ui (ui),
    _selectedButton (0) {
  driver->refreshPorts(*ui);
  setReadPorts (driver->getReadPorts (*ui));
  setWritePorts (driver->getWritePorts (*ui));
  driver->refreshSubscriptions(*ui);
  setSubscriptions (driver->getSubscriptions (*ui));

  bool(ButtonArray::* redrawFunc)(GdkEventExpose *) = &ButtonArray::redrawSubscriptions;
  _drawingArea.signal_expose_event().connect (SigC::slot (*this, redrawFunc));

  pack_start (_readBox);
  pack_start (_drawingArea);
  pack_start (_writeBox);

  show_all ();
}


void ButtonArray::refreshArray () {
  try {
    _driver->refreshPorts(*_ui);
  } catch (Exception e) {
     Gtk::MessageDialog md (e.desc, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
     md.run ();
  }

  setReadPorts (_driver->getReadPorts(*_ui));
  setWritePorts (_driver->getWritePorts (*_ui));

  try {
    _driver->refreshSubscriptions(*_ui);
  } catch (Exception e) {
     Gtk::MessageDialog md (e.desc, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
     md.run ();
  }

  setSubscriptions (_driver->getSubscriptions (*_ui));
  redrawSubscriptions ();
  _drawingArea.queue_draw();
}

void ButtonArray::setReadPorts (const list<Addr *>& ports) {
  setPorts (_readBox, _readButtons, ports, true);
}

void ButtonArray::setWritePorts (const list<Addr *>& ports) {
  setPorts (_writeBox, _writeButtons, ports, false);
}

void ButtonArray::setPorts (Gtk::VBox& buttonBox,
                            list<PortButton *>& buttonList,
                            const list<Addr *>& ports,
                            bool read) {
  for (list<PortButton *>::iterator i = buttonList.begin();
       i != buttonList.end();
       ++i) {
    buttonBox.remove (**i);
  }
  buttonList.clear ();
  
  guint index = 0;
  for (list<Addr *>::const_iterator i = ports.begin();
       i != ports.end();
       ++i) {
    PortButton * portButton = manage (new PortButton (*i, this, read, index++));
    buttonBox.pack_start (*portButton, Gtk::FILL);
    buttonList.push_back (portButton);
  }

  show_all ();
}

void ButtonArray::setSubscriptions (const list<Subscription>& subscriptions) {
  _subscriptions.clear ();
  _subscriptions = subscriptions;
}

const PortButton * ButtonArray::findButton
  (const list<PortButton *>& buttonList, const Addr * addr) {
  for (list<PortButton *>::const_iterator i = buttonList.begin();
       i != buttonList.end();
       ++i) {
    if (!*i) {
      cerr << DEBUG_STRING << "portbutton is null!" << endl;
    }
    if (!(*i)->addr()) {
      cerr << DEBUG_STRING << "portbutton's addr is null!" << endl;
    }
    const Addr * buttonAddr = (*i)->addr();
    bool t = buttonAddr->equals (addr);
    if ((*i)->addr()->equals (addr)) {
      return (*i);
    }
  }
  return 0;
}

bool ButtonArray::redrawSubscriptions () {
  GtkAllocation drawAlloc (_drawingArea.get_allocation());
  Glib::RefPtr<Gdk::Window> window = _drawingArea.get_window();
  if (window.is_null()) {
    cerr << DEBUG_STRING << "window == null; returning" << endl;
    return false;
  }
  Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create (window);

  int cw, ch;
  window->get_size (cw, ch);
  window->draw_rectangle (_drawingArea.get_style()->get_bg_gc(Gtk::STATE_NORMAL), 1,
                          0, 0, cw, ch);

  for (list<Subscription>::const_iterator i = _subscriptions.begin();
       i != _subscriptions.end();
       ++i) {
    PortButton * readButton = (PortButton *) findReadButton (i->from());
    if (!readButton) {
      cerr << DEBUG_STRING << "readbutton for " << i->from()->getName() << " not found!" << endl;
      continue;
    }
    PortButton * writeButton = (PortButton *) findWriteButton (i->to());
    if (!writeButton) {
      cerr << DEBUG_STRING << "writebutton for " << i->to()->getName() << " not found!" << endl;
      continue;
    }
    GtkAllocation readAlloc (readButton->get_allocation());
    GtkAllocation writeAlloc (writeButton->get_allocation());
    int bx = 0;
    int by = readAlloc.y + readAlloc.height / 2;
    int ex = drawAlloc.width;
    int ey = writeAlloc.y + writeAlloc.height / 2;
    Colour& colour = getColour (readButton->index());
    _drawingArea.get_colormap()->alloc_color (colour);
    gc->set_foreground(colour);
    window->draw_line (gc, bx, by, ex, ey);
  }

  return true;
}

void ButtonArray::readButtonClicked (PortButton * portButton) {
  if (_selectedButton) {
    ((Gtk::Label*)(_selectedButton->get_child()))->set_markup
      (((Gtk::Label*)(_selectedButton->get_child()))->get_text());
    if (_selectedButton == portButton) {
      _selectedButton = 0;
      return;
    }
  }
  _selectedButton = portButton;
  ((Gtk::Label*)(_selectedButton->get_child()))->set_markup
    ("<span foreground=\"#ff0000\">" +
     ((Gtk::Label*)(_selectedButton->get_child()))->get_text()
     + "</span>");
}

void ButtonArray::writeButtonClicked (PortButton * portButton) {
  if (!_selectedButton) {
    bool did_something = false;
    for (list<Subscription>::iterator it = _subscriptions.begin();
         it != _subscriptions.end();
         ++it) {
      if (it->to()->equals(portButton->addr())) {
        try {
          _driver->unsubscribePorts (*_ui, it->from(), portButton->addr());
          did_something = true;
        } catch (Exception e) {
          _ui->log (ERROR, e.desc);
          Gtk::MessageDialog md (e.desc, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
          md.run();
          return;
        }
      }
    }

    if (did_something) {
      _driver->refreshSubscriptions (*_ui);
      setSubscriptions (_driver->getSubscriptions (*_ui));
      redrawSubscriptions ();
    }
    return;
  };

  try {
    _driver->subscribePorts (*_ui, _selectedButton->addr(), portButton->addr());
  } catch (Exception e) {
    _ui->log (ERROR, string("Could not subscribe ports:\n  ") + e.desc);
    Gtk::MessageDialog md (e.desc, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
    md.run();
    return;
  }


  _subscriptions.push_back (Subscription (_selectedButton->addr(), portButton->addr()));

  ((Gtk::Label*)(_selectedButton->get_child()))->set_markup
      (((Gtk::Label*)(_selectedButton->get_child()))->get_text());
  _selectedButton = 0;

  redrawSubscriptions();
}

/////////////////////////////////
//////// PortButton /////////////
/////////////////////////////////

void PortButton::on_clicked () {
  if (_read) {
    _buttonArray->readButtonClicked (this);
  } else {
    _buttonArray->writeButtonClicked (this);
  }
}

/////////////////////////////////
//////// MainWindow /////////////
/////////////////////////////////

void MainWindow::jfdToggled () {
  bool b = _jfdButton.get_active ();
  _driver->setJFD (b);
}

MainWindow::MainWindow (string title, APB::Driver * driver, APB::UI * ui)
  : _logBox(),
    _buttonArray (driver, ui),
    _jfdButton ("JFD"),
    _driver (driver) {

  set_title (ustring (title));
  set_default_size (550, 300);
  std::string iconFilename (std::string(PKGDATADIR) + "/5-pin-din.png");
  try {
    set_icon (Gdk::Pixbuf::create_from_file (iconFilename));
  } catch (Glib::FileError e) {
    cerr << DEBUG_STRING << "could not load icon " << iconFilename
         << ": " << strerror (e.code()) << endl;
  } catch (Gdk::PixbufError e) {
    cerr << DEBUG_STRING << "could not load icon " << iconFilename
         << " due to dodgey image data" << endl;
  }

  Gtk::Tooltips * tooltips = manage (new Gtk::Tooltips());

  Gtk::VBox * vbox = manage (new Gtk::VBox());
  Gtk::VPaned * vpane = manage (new Gtk::VPaned());

  Gtk::ScrolledWindow * scrollWindow =
    manage (new Gtk::ScrolledWindow ());
  scrollWindow->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);

  Gtk::Adjustment * ha = manage (new Gtk::Adjustment (0.5, 0.0, 1.0));
  Gtk::Adjustment * va = manage (new Gtk::Adjustment (0.5, 0.0, 1.0));
  Gtk::Viewport * viewport = manage (new Gtk::Viewport(*ha, *va));
  Gtk::HBox * topButtonBox = manage (new Gtk::HBox());

  Gtk::Button * quitButton = manage (new Gtk::Button (ustring("Quit")));
  quitButton->unset_flags (Gtk::CAN_FOCUS);
  quitButton->signal_clicked().connect (SigC::slot(*(SigC::Object *)ui, &UI::stop));

  Gtk::Button * refreshButton = manage (new Gtk::Button (ustring("Refresh")));
  refreshButton->unset_flags (Gtk::CAN_FOCUS);
  refreshButton->signal_clicked().connect (SigC::slot(_buttonArray, &ButtonArray::refreshArray));

  _jfdButton.set_active (true);
  _jfdButton.unset_flags (Gtk::CAN_FOCUS);
  _jfdButton.signal_clicked().connect (SigC::slot(*this, &MainWindow::jfdToggled));
  tooltips->set_tip (_jfdButton, "uncheck this to be asked for subscription options");

  Gtk::ScrolledWindow * logScroll = 
    manage (new Gtk::ScrolledWindow ());
  logScroll->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);

  _logBox.set_editable (false);
  _logBox.set_wrap_mode (Gtk::WRAP_WORD);
  _logBox.unset_flags (Gtk::CAN_FOCUS);

  viewport->add (_buttonArray);
  scrollWindow->add (*viewport);
  logScroll->add (_logBox);
  vpane->pack2 (*logScroll, Gtk::SHRINK);
  vpane->pack1 (*scrollWindow, Gtk::EXPAND);
  topButtonBox->pack_start (*refreshButton, Gtk::FILL);
  topButtonBox->pack_start (*quitButton, Gtk::FILL);
  topButtonBox->pack_start (_jfdButton, Gtk::FILL);
  vbox->pack_start (*topButtonBox, Gtk::FILL);
  vbox->pack_start (*vpane);
  add (*vbox);
  show_all ();
}

////////////////////////////////
/////// ChoiceButton ///////////
////////////////////////////////

/*ChoiceButton::ChoiceButton (ChoiceDialog * dialog)
  : Gtk::Button (ustring ("OK")),
    _dialog (dialog),
    _response (0) {
}

void ChoiceButton::on_clicked () {
  _dialog->response (_response);
}


////////////////////////////////
/////// ChoiceDialog ///////////
////////////////////////////////

ChoiceDialog::ChoiceDialog (const string& title,
                            const string& message,
                            const list<pair<string, int> >& choices,
                            Gtk::Window& parent)
  : Gtk::Dialog (ustring (title), parent, false),
    _choices (&choices), 
    _response (0) {
  MainWindow * mw = (MainWindow *) &parent;
  mw->log (DEBUG, DEBUG_STRING + "creating ok button");
  ChoiceButton * choiceButton = manage (new ChoiceButton (this));

  Gtk::VBox * vbox = get_vbox ();
  mw->log (DEBUG, DEBUG_STRING + "packing ok button");
  vbox->pack_end (*choiceButton);
  mw->log (DEBUG, DEBUG_STRING + "packing label button");
  Gtk::Label * messageLabel = manage (new Gtk::Label (ustring(message)));
  vbox->pack_start (*messageLabel);
  
  for (list<pair<string, int> >::const_iterator it = choices.begin();
       it != choices.end();
       ++it) {
    mw->log (DEBUG, DEBUG_STRING + "packing check button " + it->first);
    Gtk::CheckButton * checkButton = manage (new ChoiceCheckButton (it->first, it->second, choiceButton));
    vbox->pack_start (*checkButton);
  }

  vbox->show_all ();
}

ChoiceDialog::~ChoiceDialog () {
}*/


//////////////////////
/////// UI ///////////
//////////////////////

UI::UI (string title, int argc, char ** argv, void * gtk, APB::Driver * driver)
  : _gtk (gtk ? (Gtk::Main *) gtk : new Gtk::Main (argc, argv)),
    _driver (driver),
    _mainWindow (title, driver, this),
    _title (title) {
  if (!_gtk) {
    string s (DEBUG_STRING + "failed to initialise Gtk--");
    throw APB::Exception (s);
  }
}  


} /* namespace GTKmm */
} /* namespace APB */
